# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import Callable


def catch_exceptions(func: Callable) -> Callable[..., bool]:
    logger = logging.getLogger(__name__)

    def callable_function(*args, **kwargs) -> bool:
        try:
            func(*args, **kwargs)
            return True
        except BaseException as e:
            logger.exception("An exception occurred: %s", e)
            return False

    return callable_function



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import Dict, List, Tuple
import os
from datetime import datetime
from pandas import DataFrame


USERS_ROOT_DIRECTORY = ["users"]
SECTIONS_ROOT_DIRECTORY = ["sections"]
SECTION_ASSOCIATIONS_ROOT_DIRECTORY = ["section={id}", "section-associations"]
ASSIGNMENT_ROOT_DIRECTORY = ["section={id}", "assignments"]
SUBMISSION_ROOT_DIRECTORY = ["section={id1}", "assignment={id2}", "submissions"]
GRADES_ROOT_DIRECTORY = ["section={id}", "grades"]
SECTION_ACTIVITY_DIRECTORY = ["section={id}", "section-activities"]
SYSTEM_ACTIVITY_ROOT_DIRECTORY = ["system-activities"]

logger = logging.getLogger(__name__)


def _normalized_directory_template(
    output_directory: str, additional_path: List[str]
) -> str:
    return os.path.join(os.path.normpath(output_directory), *additional_path)


def _write_csv(df_to_write: DataFrame, output_date: datetime, directory: str):
    """
    Write a LMS UDM DataFrame to a CSV file

    Parameters
    ----------
    df_to_write: DataFrame
        is a LMS UDM DataFrame
    output_date: datetime
        is the timestamp for the filename
    directory: str
        is the directory the file will go in
    """
    filename: str = output_date.strftime("%Y-%m-%d-%H-%M-%S")
    path = os.path.join(directory, f"{filename}.csv")

    try:
        os.makedirs(directory, exist_ok=True)
        df_to_write.to_csv(path, index=False)

        logger.info(f"Generated file => {path}")
    except Exception:
        logger.exception("An exception occurred while writing file %s", path)


def _write_multi_csv(
    dfs_to_write: Dict[str, DataFrame], output_date: datetime, directory_template: str
):
    """
    Write a series of LMS UDM DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[str, DataFrame]
        is a Dict of id/LMS UDM DataFrame pairs
    output_date: datetime
        is the timestamp for the filename
    directory_template: str
        is the directory the file will go in, with an {id} placeholder
    """
    assert "{id}" in directory_template

    for id_placeholder, df_to_write in dfs_to_write.items():
        directory: str = directory_template.format(id=id_placeholder)
        _write_csv(df_to_write, output_date, directory)


def _write_multi_tuple_csv(
    dfs_to_write: Dict[Tuple[str, str], DataFrame],
    output_date: datetime,
    directory_template: str,
):
    """
    Write a series of LMS UDM DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[Tuple[str, str], DataFrame]
        is a Dict of 2 id tuples/LMS UDM DataFrame pairs
    output_date: datetime
        is the timestamp for the filename
    directory_template: str
        is the directory the file will go in, with an {id1} and an {id2} placeholder
    """
    assert "{id1}" in directory_template
    assert "{id2}" in directory_template

    for id_tuple, df_to_write in dfs_to_write.items():
        (id1, id2) = id_tuple
        directory: str = directory_template.format(id1=id1, id2=id2)
        _write_csv(df_to_write, output_date, directory)


def _fill_in_missing_section_ids(
    dfs_to_write: Dict[str, DataFrame], all_section_ids: List[str]
) -> Dict[str, DataFrame]:
    """
    Fill in any missing DataFrames for a Section with an empty DataFrame

    Parameters
    ----------
    dfs_to_write: Dict[str, DataFrame]
        is a Dict of section id/DataFrame pairs
    all_section_ids: List[str]
        is a list of all known section ids


    Returns
    -------
    Dict[str, DataFrame]
        a new Dict of section id/DataFrame pairs, with missing DataFrames
        for a Section filled in
    """
    full_dfs_to_write = dfs_to_write.copy()
    for section_id in all_section_ids:
        if section_id not in full_dfs_to_write:
            full_dfs_to_write[section_id] = DataFrame()
    return full_dfs_to_write


def write_users(df_to_write: DataFrame, output_date: datetime, output_directory: str):
    """
    Write a LMS UDM Users DataFrame to a CSV file

    Parameters
    ----------
    df_to_write: DataFrame
        is a LMS UDM Users DataFrame
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_csv(
        df_to_write,
        output_date,
        _normalized_directory_template(output_directory, USERS_ROOT_DIRECTORY),
    )


def write_sections(
    df_to_write: DataFrame, output_date: datetime, output_directory: str
):
    """
    Write a LMS UDM Sections DataFrame to a CSV file

    Parameters
    ----------
    df_to_write: DataFrame
        is a LMS UDM Sections DataFrame
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_csv(
        df_to_write,
        output_date,
        _normalized_directory_template(output_directory, SECTIONS_ROOT_DIRECTORY),
    )


def write_section_associations(
    dfs_to_write: Dict[str, DataFrame],
    all_section_ids: List[str],
    output_date: datetime,
    output_directory: str,
):
    """
    Write a series of LMS UDM UserSectionAssociation DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[str, DataFrame]
        is a Dict of id/LMS UDM UserSectionAssociation DataFrame pairs
    all_section_ids: List[str]
        is a list of all known section ids
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_multi_csv(
        _fill_in_missing_section_ids(dfs_to_write, all_section_ids),
        output_date,
        _normalized_directory_template(
            output_directory, SECTION_ASSOCIATIONS_ROOT_DIRECTORY
        ),
    )


def write_assignments(
    dfs_to_write: Dict[str, DataFrame],
    all_section_ids: List[str],
    output_date: datetime,
    output_directory: str,
):
    """
    Write a series of LMS UDM Assignments DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[str, DataFrame]
        is a Dict of section id/LMS UDM Assignments DataFrame pairs
    all_section_ids: List[str]
        is a list of all known section ids
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """

    _write_multi_csv(
        _fill_in_missing_section_ids(dfs_to_write, all_section_ids),
        output_date,
        _normalized_directory_template(output_directory, ASSIGNMENT_ROOT_DIRECTORY),
    )


def write_assignment_submissions(
    dfs_to_write: Dict[Tuple[str, str], DataFrame],
    output_date: datetime,
    output_directory: str,
):
    """
    Write a series of LMS UDM AssignmentSubmissions DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[Tuple[str, str], DataFrame]
        is a Dict of 2 id tuples/LMS UDM AssignmentSubmissions DataFrame pairs
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_multi_tuple_csv(
        dfs_to_write,
        output_date,
        _normalized_directory_template(output_directory, SUBMISSION_ROOT_DIRECTORY),
    )


def write_grades(
    dfs_to_write: Dict[str, DataFrame],
    all_section_ids: List[str],
    output_date: datetime,
    output_directory: str,
):
    """
    Write a series of LMS UDM Grades DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[str, DataFrame]
        is a Dict of section id/LMS UDM Grades DataFrame pairs
    all_section_ids: List[str]
        is a list of all known section ids
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_multi_csv(
        _fill_in_missing_section_ids(dfs_to_write, all_section_ids),
        output_date,
        _normalized_directory_template(output_directory, GRADES_ROOT_DIRECTORY),
    )


def write_section_activities(
    dfs_to_write: Dict[str, DataFrame],
    all_section_ids: List[str],
    output_date: datetime,
    output_directory: str,
):
    """
    Write a series of LMS UDM SectionActivity DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: Dict[str, DataFrame]
        is a Dict of section id/LMS UDM SectionActivity DataFrame pairs
    all_section_ids: List[str]
        is a list of all known section ids
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_multi_csv(
        _fill_in_missing_section_ids(dfs_to_write, all_section_ids),
        output_date,
        _normalized_directory_template(output_directory, SECTION_ACTIVITY_DIRECTORY),
    )


def write_system_activities(
    df_to_write: DataFrame, output_date: datetime, output_directory: str
):
    """
    Write a series of LMS UDM System Activities DataFrames to CSV files

    Parameters
    ----------
    dfs_to_write: DataFrame
        Dataframe to write
    output_date: datetime
        is the timestamp for the filename
    output_directory: str
        is the root output directory
    """
    _write_csv(
        df_to_write,
        output_date,
        _normalized_directory_template(
            output_directory,
            [
                *SYSTEM_ACTIVITY_ROOT_DIRECTORY,
                f"date={output_date.strftime('%Y-%m-%d')}",
            ],
        ),
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from datetime import datetime
from typing import List
from pandas import DataFrame, Series, read_sql_query, to_datetime
import sqlalchemy
import xxhash

logger = logging.getLogger(__name__)

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

SYNC_COLUMNS = [
    "SourceId",
    "Json",
    "Hash",
    "CreateDate",
    "LastModifiedDate",
    "SyncNeeded",
]

SYNC_COLUMNS_SQL = """
    SourceId TEXT,
    Json TEXT,
    Hash TEXT,
    CreateDate DATETIME,
    LastModifiedDate DATETIME,
    SyncNeeded BIGINT,
    PRIMARY KEY (SourceId)
    """


def _json_hash_encode(row: Series) -> Series:
    """
    Take a DataFrame row, add serialized JSON and hash

    Parameters
    ----------
    row: Series
        a DataFrame row

    Returns
    -------
    Series
        the row with the json and hash columns added
    """
    json = row.to_json()
    row["Json"] = json
    row["Hash"] = xxhash.xxh64_hexdigest(json.encode("utf-8"))
    return row


def add_hash_and_json_to(df: DataFrame) -> DataFrame:
    """
    Create Hash and Json columns for DataFrame.  Do this
    before adding any other columns e.g. SourceId

    Parameters
    ----------
    df: DataFrame
        a DataFrame with fetched data

    Returns
    -------
    DataFrame
        a new DataFrame with the json and hash columns added
    """
    return df.apply(_json_hash_encode, axis=1)


def add_sourceid_to(df: DataFrame, identity_columns: List[str]):
    """
    Create SourceId column for DataFrame with given identity columns.

    Parameters
    ----------
    df: DataFrame
        a DataFrame with fetched data
    identity_columns: List[str]
        a List of the identity columns for the resource dataframe
    """
    assert (
        Series(identity_columns).isin(df.columns).all()
    ), "Identity columns missing from dataframe"

    df[identity_columns] = df[identity_columns].astype("string")
    df["SourceId"] = df[sorted(identity_columns)].agg("-".join, axis=1)


def _create_sync_table_from_resource_df(
    resource_df: DataFrame,
    identity_columns: List[str],
    resource_name: str,
    sync_db: sqlalchemy.engine.base.Engine,
):
    """
    Take fetched data and push to a new temporary sync table.  Includes
    hash and tentative extractor CreateDate/LastModifiedDates.

    Parameters
    ----------
    resource_df: DataFrame
        a DataFrame with current fetched data.
    identity_columns: List[str]
        a List of the identity columns for the resource dataframe.
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections
    """
    with sync_db.connect() as con:
        # ensure sync table exists, need column ordering to be identical to regular table
        con.execute(f"DROP TABLE IF EXISTS Sync_{resource_name}")
        con.execute(
            f"""
            CREATE TABLE IF NOT EXISTS Sync_{resource_name} (
                {SYNC_COLUMNS_SQL}
            )
            """
        )

    sync_df: DataFrame = resource_df.copy()
    sync_df = add_hash_and_json_to(sync_df)

    # add (possibly composite) primary key, sorting for consistent ordering
    add_sourceid_to(sync_df, identity_columns)

    now: datetime = datetime.now()
    sync_df["CreateDate"] = now
    sync_df["LastModifiedDate"] = now
    sync_df["SyncNeeded"] = 1

    sync_df = sync_df[SYNC_COLUMNS]
    sync_df.set_index("SourceId", inplace=True)
    # push to temporary sync table
    sync_df.to_sql(
        f"Sync_{resource_name}", sync_db, if_exists="append", index=True, chunksize=1000
    )


def _ensure_main_table_exists(
    resource_name: str,
    con: sqlalchemy.engine.base.Connection,
):
    """
    Ensure the main resource table exists, creating if necessary.

    Parameters
    ----------
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    table_columns_sql: str
        the columns for the resource in the database, in SQL table creation form,
            with dangling commas
    con: sqlalchemy.engine.base.Connection
        an open database connection, which will not be closed by this function
    """
    con.execute(f"DROP INDEX IF EXISTS SYNCNEEDED_{resource_name}")
    con.execute(
        f"""
        CREATE TABLE IF NOT EXISTS {resource_name} (
            {SYNC_COLUMNS_SQL}
        )
        """
    )
    con.execute(
        f"CREATE INDEX IF NOT EXISTS SYNCNEEDED_{resource_name} ON {resource_name}(SyncNeeded)"
    )


def _create_unmatched_records_temp_table(
    resource_name: str,
    con: sqlalchemy.engine.base.Connection,
):
    """
    Select unmatched records into new temp table - differing by hash for same identity.
    Single entry in result set if identity only exists in one table (meaning add or missing),
        so SyncNeeded flag will indicate which table it's from.
    Double entry in result set if identity exists in both (meaning update needed),
        so SyncNeeded will show which row is from which table.

    Parameters
    ----------
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    con: sqlalchemy.engine.base.Connection
        an open database connection, which will not be closed by this function
    """
    con.execute(f"DROP INDEX IF EXISTS ID_{resource_name}")
    con.execute(f"DROP TABLE IF EXISTS Unmatched_{resource_name}")
    con.execute(
        f"""
        CREATE TABLE Unmatched_{resource_name} AS
        SELECT * FROM (
            SELECT * FROM {resource_name}
            UNION ALL
            SELECT * FROM Sync_{resource_name}
        )
        GROUP BY SourceId, Hash
        HAVING COUNT(*) = 1
        """
    )
    con.execute(
        f"CREATE INDEX IF NOT EXISTS ID_{resource_name} ON Unmatched_{resource_name}(SourceId)"
    )


def _get_true_create_dates_for_unmatched_records(
    resource_name: str,
    con: sqlalchemy.engine.base.Connection,
):
    """
    All rows start with CreateDate and LastModifiedDate initialized to "now",
        but updated rows need the original CreateDate pulled from existing table.

    Note: UPDATE-FROM is not available in sqlite until v3.33.0, thus the
        double select goofiness.

    Parameters
    ----------
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    con: sqlalchemy.engine.base.Connection
        an open database connection, which will not be closed by this function
    """
    con.execute(
        f"""
        UPDATE Unmatched_{resource_name}
            SET CreateDate = (
                SELECT c.CreateDate
                FROM {resource_name} c
                WHERE c.SourceId = Unmatched_{resource_name}.SourceId
            )
            WHERE EXISTS (
                SELECT *
                FROM {resource_name} c
                WHERE c.SourceId = Unmatched_{resource_name}.SourceId
            ) AND SyncNeeded = 1
        """
    )


def _update_resource_table_with_changes(
    resource_name: str,
    con: sqlalchemy.engine.base.Connection,
):
    """
    Update main resource table with new and updated records

    Parameters
    ----------
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    primary_keys: str
        a comma separated list of the primary key columns for the resource,
        e.g. "id,courseId"
    con: sqlalchemy.engine.base.Connection
        an open database connection, which will not be closed by this function
    """
    CHANGED_ROWS_CTE = f"""
                        changedRows AS (
                            SELECT * FROM Unmatched_{resource_name}
                            WHERE (SourceId) IN (
                                SELECT SourceId FROM Unmatched_{resource_name}
                                GROUP BY SourceId
                                HAVING COUNT(*) > 1
                            ) AND SyncNeeded = 1
                        )
                        """

    # delete obsolete data from regular table
    con.execute(
        # changed rows CTE (from SyncNeeded side only)
        f"""
        WITH
        {CHANGED_ROWS_CTE}
        DELETE FROM {resource_name}
        WHERE (SourceId) IN (
            SELECT SourceId from changedRows
        )
        """
    )

    # insert new and changed data into regular table
    con.execute(
        #    changed rows CTE (from SyncNeeded side only)
        #    new rows CTE (also from SyncNeeded side)
        f"""
        WITH
            {CHANGED_ROWS_CTE},
            newRows AS (
                SELECT * FROM Unmatched_{resource_name}
                WHERE (SourceId) IN (
                    SELECT SourceId FROM Unmatched_{resource_name}
                    GROUP BY SourceId
                    HAVING COUNT(*) = 1 AND SyncNeeded = 1
                )
            )
        INSERT INTO {resource_name}
            SELECT * FROM Unmatched_{resource_name}
            WHERE (SourceId) IN (
                SELECT SourceId FROM changedRows
                UNION ALL
                SELECT SourceId FROM newRows
            ) AND SyncNeeded = 1
        """
    )

    con.execute(
        # reset SyncNeeded flag on main table
        f"""
        UPDATE {resource_name}
        SET SyncNeeded = 0
        WHERE SyncNeeded != 0
        """
    )


def _update_dataframe_with_true_dates(
    resource_df: DataFrame,
    identity_columns: List[str],
    resource_name: str,
    con: sqlalchemy.engine.base.Connection,
) -> DataFrame:
    """
    Update main resource DataFrame with reconciled CreateDate/LastModifiedDates

    Parameters
    ----------
    resource_df: DataFrame
        an API DataFrame with current fetched data which
        will be mutated by updating CreateDate/LastModifiedDate
    identity_columns: List[str]
        a List of the identity columns for the resource dataframe.
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    primary_keys: str
        a comma separated list of the primary key columns for the resource,
        e.g. "id,courseId"
    con: sqlalchemy.engine.base.Connection
        an open database connection, which will not be closed by this function

    Returns
    -------
    DataFrame
        a DataFrame with reconciled CreateDate/LastModifiedDate
    """
    assert (
        Series(identity_columns).isin(resource_df.columns).all()
    ), "Identity columns missing from dataframe"

    # fetch DataFrame with reconciled CreateDate/LastModifiedDate for sync records
    update_dates_query = f"""
                            SELECT SourceId, CreateDate, LastModifiedDate
                            FROM {resource_name}
                            WHERE (SourceId) IN (
                                SELECT SourceId FROM Sync_{resource_name}
                            )
                            """
    create_date_df = read_sql_query(update_dates_query, con)
    create_date_df["SourceId"] = create_date_df["SourceId"].astype("string")

    add_sourceid_to(resource_df, identity_columns)
    resource_df["SourceId"] = resource_df["SourceId"].astype("string")

    result_df = resource_df.join(create_date_df.set_index("SourceId"), on="SourceId")

    # reset index so no columns are hidden
    result_df.drop(["SourceId"], axis=1, inplace=True)

    # convert dates to string format
    result_df["CreateDate"] = to_datetime(result_df["CreateDate"]).dt.strftime(DATE_FORMAT)
    result_df["LastModifiedDate"] = to_datetime(result_df["LastModifiedDate"]).dt.strftime(DATE_FORMAT)

    return result_df


def sync_to_db_without_cleanup(
    resource_df: DataFrame,
    identity_columns: List[str],
    resource_name: str,
    sync_db: sqlalchemy.engine.base.Engine,
):
    """
    Take fetched data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Does not delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a DataFrame with current fetched data
    identity_columns: List[str]
        a List of the identity columns for the resource dataframe.
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    assert (
        Series(identity_columns).isin(resource_df.columns).all()
    ), "Identity columns missing from dataframe"

    _create_sync_table_from_resource_df(
        resource_df, identity_columns, resource_name, sync_db
    )

    with sync_db.connect() as con:
        _ensure_main_table_exists(resource_name, con)
        _create_unmatched_records_temp_table(resource_name, con)
        _get_true_create_dates_for_unmatched_records(resource_name, con)
        _update_resource_table_with_changes(resource_name, con)
        result_df: DataFrame = _update_dataframe_with_true_dates(
            resource_df, identity_columns, resource_name, con
        )

    return result_df


def cleanup_after_sync(resource_name: str, sync_db: sqlalchemy.engine.base.Engine):
    """
    Delete sync temporary tables if they exist

    Parameters
    ----------
    resource_name: str
        the name of the API resource, e.g. "Courses", to be used in SQL
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections
    """
    with sync_db.connect() as con:
        con.execute(f"DROP TABLE IF EXISTS Sync_{resource_name}")
        con.execute(f"DROP TABLE IF EXISTS Unmatched_{resource_name}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edfi_lms_extractor_lib',
 'edfi_lms_extractor_lib.api',
 'edfi_lms_extractor_lib.csv_generation',
 'edfi_lms_extractor_lib.helpers']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.3.19,<2.0.0', 'pandas>=1.1.1,<2.0.0', 'xxhash>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'edfi-lms-extractor-lib',
    'version': '1.0.0',
    'description': 'Shared functions library for Ed-Fi LMS Extractor projects',
    'long_description': '# Extractor Shared Library\n\nShared library for use in the [Ed-Fi LMS\nToolkit](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit).\n\n## Legal Information\n\nCopyright (c) 2021 Ed-Fi Alliance, LLC and contributors.\n\nLicensed under the [Apache License, Version\n2.0](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/LICENSE) (the\n"License").\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.\n\nSee\n[NOTICES](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/NOTICES.md)\nfor additional copyright and license notifications.\n',
    'author': 'Ed-Fi Alliance, LLC, and contributors',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://techdocs.ed-fi.org/display/EDFITOOLS/LMS+Toolkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
